from .agent import Agent 
from .trainer import QTrainer
from .models import LinearQNet
from .utils import (
    create_session, 
    create_folder_struct,
    delete_session, 
    add_dict_inside_session
)

from .config import CONFIG 
from typing import Callable, Optional
from rewards_envs import CarGame 

class RewardsWorkflow:
    def __init__(self, session_id : str, session_root_folder : Optional[str] = None) -> None:
        self.session_id = session_id
        self.session_root_folder = session_root_folder
        self.session_root_dir = create_folder_struct(
            dir=session_root_folder, 
            name = CONFIG['REWARDS_PARENT_CONFIG_DIR']
        )
        
        create_session(
            session_name = session_id, 
            session_root_dir=self.session_root_dir
        )

        self.env_config = None
        self.agent_config = None
        self.training_config = None 
    
    def register_env_parameters(self, env_config):
        env_config_dict = env_config.__dict__.copy() 
        add_dict_inside_session(
            session_id = self.session_id,
            configuration = env_config_dict, 
            config_name=CONFIG['REWARDS_CONFIG_ENV_FILE_NAME'], 
            session_root_folder=self.session_root_folder, 
            rewrite = True, multi_config=False 
        )
        self.env_config = env_config
        
    def register_agent_parameters(self, agent_config):
        agent_config_dict = agent_config.__dict__.copy() 
        add_dict_inside_session(
            session_id = self.session_id, 
            configuration=agent_config_dict, 
            session_root_folder=self.session_root_folder, 
            config_name=CONFIG['REWARDS_CONFIG_AGENT_FILE_NAME'],
            rewrite=True, multi_config=False 
        )
        self.agent_config = agent_config
    
    def register_training__parameters(self, training_tracks : int,  evaluation_tracks : int, num_episodes : int):
        training_params = {
            'traning_tracks' : training_tracks, 
            'evaluation_tracks' : evaluation_tracks, 
            'num_episodes' : num_episodes
        } 
        
        add_dict_inside_session(
            session_id = self.session_id, 
            configuration=training_params, 
            session_root_folder=self.session_root_dir, 
            config_name=CONFIG['REWARDS_CONFIG_TRAINING_FILE_NAME'], 
            rewrite=True, multi_config=False
        )

    def register_reward_function(self, reward_function : Callable, reward_function_name : str):
        ...  
    
    def train(self, env_track : int):
        raise NotImplementedError
    
    def evaluate(self, env_track : int, use_train_track : bool = False):
        raise NotImplementedError

    @property
    def delete_session(self):
        delete_session(
            session_id=self.session_id, 
            session_root_dir=self.session_root_folder
        ) 
        print("=> Session Deleted successfully")

    def push_to_s3(self, aws_session_id, aws_folder_id):
        """
        WORKFLOW:
        --------
        - First it will calculate the rewards for all the training envs 
        - Then it will calculate the rewards for all the evaluation envs 
        - Finally it will calculate all other factors from:
            - agent_params (neural network size)
            - time taken to complete the training 
        package all the information to a json and the model.pth 
        and send to AWS S3 
        """
        raise NotImplementedError
    


import os 
import torch 
import random 
import numpy as np 
import torch.nn as nn 
import torch.optim as optim 
from typing import Any, List, Union

from .agent import Agent 
from .utils import RootMeanSquaredError, MeanAbsoluteError

class QTrainer(Agent):
    def __init__(self, **training_params):
        self.lr = training_params["lr"]
        self.gamma = training_params["gamma"]
        self.epsilon = training_params["epsilon"]

        print(self.epsilon)
        
        self.model = training_params["model"]
        loss_fn, optimizer_info = self._get_loss_optimizer_info(
            training_params["loss"], training_params["optimizer"]
        )
        self.criterion = loss_fn()
        self.optimizer = optimizer_info(self.model.parameters(), lr=self.lr)

        super(QTrainer, self).__init__(
            model=self.model,
            lr=self.lr,
            epsilon=self.epsilon,
            gamma=self.gamma,
            checkpoint_folder_path=training_params['checkpoint_folder_path'], 
            model_name=training_params['model_name']
        )
        

    def _get_loss_optimizer_info(
        self, loss: str, optimizer: str
    ) -> List[Union[int, str, float]]:
        """_summary_

        Args:
            loss (str): _description_
            optimizer (str): _description_

        Returns:
            List[str, Union[int, str, float]]: _description_
        """
        loss_info = {
            "mse": torch.nn.MSELoss,
            "rmse": RootMeanSquaredError,
            "mae": MeanAbsoluteError,
        }

        optimizer_info = {
            "adam": optim.Adam,
            "rmsprop": optim.RMSprop,
            "adagrad": optim.Adagrad,
        }

        return loss_info[loss], optimizer_info[optimizer]

    def step(
        self,
        state: Any,
        action: Union[np.ndarray, List[Union[float, int]]],
        reward: Union[float, int],
        next_state: Any,
        done: bool,
    ) -> None:
        """Single step function for a single episode for the agent's training

        Args:
            state (Any): The current state of the environment
            action (Union[np.ndarray, List[Union[float, int]]]): The action taken by the agent
            reward (Union[float, int]): The reward that the agent gets
            next_state (Any): Next state after the action taken by the agent
            done (bool): Whether the game terminates or not
        """
        state = torch.tensor(state, dtype=torch.float32)
        next_state = torch.tensor(next_state, dtype=torch.float32)
        action = torch.tensor(action, dtype=torch.float32)
        reward = torch.tensor(reward, dtype=torch.float32)

        if len(state.shape) == 1:
            state = torch.unsqueeze(state, dim=0)
            next_state = torch.unsqueeze(next_state, dim=0)
            action = torch.unsqueeze(action, dim=0)
            reward = torch.unsqueeze(reward, dim=0)
            done = (done,)

        state_prediction = self.model(state)
        state_target = (
            state_prediction.clone()
        )  # TODO: Why target is the same as prediction?

        for idx in range(len(done)):
            Q_new = reward[idx]
            if not done[idx]:
                Q_new = reward[idx] + self.gamma * torch.max(
                    self.model(next_state[idx])
                )
            state_target[idx][torch.argmax(action[idx]).item()] = Q_new

        self.optimizer.zero_grad()
        loss = self.criterion(state_target, state_prediction)
        loss.backward()
        self.optimizer.step()
        
        
    def train_long_memory(self) -> None:
        """Trains the agent for a longer step saving state-actions to the memory
     print("stepping")
        Returns:
            None
        """
        if len(self.memory) > self.BATCH_SIZE:
            mini_sample = random.sample(self.memory, self.BATCH_SIZE)
        else:
            mini_sample = self.memory
        
        # what this line is doing?
        states, actions, rewards, next_states, dones = zip(*mini_sample)
        self.step(states, actions, rewards, next_states, dones)
    
    def train_short_memory(self, state, action, reward, next_state, done):
        """Trains the agent for a single step without any memory search

        Args:
            state (_type_): _description_
            action (_type_): _description_
            reward (_type_): _description_
            next_state (_type_): _description_
            done (function): _description_

        Returns:
            _type_: _description_
        """
        return self.step(state, action, reward, next_state, done)
    
    def train_step(self, game):
        state_old = self.get_state(game)
        final_move = self.get_action(state_old, mode="training")
        
        if game.config.render_mode == "rgb_array":
            reward, done, score, pixel_data = game.play_step(final_move)
        else:
            reward, done, score = game.play_step(final_move)
            
        state_new = self.get_state(game)
        self.train_short_memory(state_old, final_move, reward, state_new, done)
        self.remember(state_old, final_move, reward, state_new, done)
        return [reward, done, score, pixel_data] if game.config.render_mode == "rgb_array" else [reward, done, score]
    
    def evaluate(self, game):
        state_old = self.get_state(game)
        final_move = self.get_action(state_old, mode="evaluation")
        if game.config.render_mode == "rgb_array":
            reward, done, score, pixel_data = game.play_step(final_move)
        else:
            reward, done, score = game.play_step(final_move)
        return [reward, done, score, pixel_data] if game.config.render_mode == "rgb_array" else [reward, done, score]


import os
import torch
from dataclasses import dataclass
from typing import Optional, Tuple, Dict, Any

@dataclass(kw_only=True)
class GlobalConfig:
    USE_CUDE_IF_AVAILABLE: bool = False
    SCREEN_SIZE: Optional[Tuple] = (800, 700)
    ENABLE_WANDB: Optional[bool] = False
    IN_FEATURES: int = 5
    REWARDS_PARENT_CONFIG_DIR: str = ".rewards_ai"
    REWARDS_CONFIG_MODEL_FOLDER_NAME: str = "session_saved_models"
    REWARDS_CONFIG_METRIC_FOLDER_NAME: str = "session_metrics"
    
    REWARDS_CONFIG_ENV_FILE_NAME : str = "environment_config"
    REWARDS_CONFIG_AGENT_FILE_NAME : str = "agent_config"
    REWARDS_CONFIG_TRAINING_FILE_NAME : str = "training_config"
    REWARDS_CONFIG_REWARD_FN_FILE_NAME : str = "reward_functions"
    

CONFIG = GlobalConfig()


import os 
import json 
import torch 
import shutil 
from  .config import CONFIG
from typing import Optional, Union, Any, Dict

class RootMeanSquaredError(torch.nn.Module):
    def __init__(self):
        """Root mean squared error function in PyTorch"""
        super(RootMeanSquaredError, self).__init__()

    def forward(self, x, y):
        return torch.sqrt(torch.mean((x - y) ** 2))


class MeanAbsoluteError(torch.nn.Module):
    def __init__(self):
        """
        Mean absolute error function in PyTorch
        """
        super(MeanAbsoluteError, self).__init__()

    def forward(self, x, y):
        return torch.mean(torch.abs(x - y))


def get_home_path():
    # get the home directory using os
    return os.path.expanduser("~")

def update_graphing_file(session_id: str, data: dict, dir: Optional[str] = None, name: Optional[str] = None) -> None:
    f = open("D:/Prototypes/rewards.ai/training-platform\src/assets/temp.json", "w")
    f.write(json.dumps(data))
    

def create_folder_struct(dir: Optional[str] = None, name: Optional[str] = None) -> None:
    """Creates a root folder to store all the session configuration

    Args:
        dir (Optional[str], optional): _description_. Defaults to None.
        name (Optional[str], optional): _description_. Defaults to None.
    """
    dir = get_home_path() if dir is None else dir
    name = CONFIG['REWARDS_PARENT_CONFIG_DIR'] if name is None else name

    if not os.path.exists(os.path.join(dir, name)):
        os.mkdir(os.path.join(dir, name))

    else:
        print("=> Folder already exists")
    
    return os.path.join(dir, name)


def create_session(session_name: str, session_root_dir: Optional[str] = None) -> Union[str, bool]:
    """Creates a folder inside the root folder to store all the configuration for a particular session

    Args:
        session_dir (Optional[str], optional): _description_. Defaults to None.
        session_name (Optional[str], optional): _description_. Defaults to None.
    """
    # TODO : Make a proper return value to detect error

    session_root_dir = (
        os.path.join(get_home_path(), CONFIG["REWARDS_PARENT_CONFIG_DIR"])
        if session_root_dir is None
        else session_root_dir
    )
    session_dir = os.path.join(session_root_dir, session_name)

    if not os.path.exists(session_dir):
        os.mkdir(session_dir)
        os.mkdir(os.path.join(session_dir, CONFIG["REWARDS_CONFIG_MODEL_FOLDER_NAME"]))

        # create a metrics.json
        null_metrics = []
        with open(os.path.join(session_dir, "metrics.json"), "w") as metrics_file:
            json.dump(null_metrics, metrics_file)


def add_inside_session(
    session_id: str, config_name: str, rewrite: bool = False, multi_config: bool = False, **kwargs
):
    """Add configuration inside a session folder

    Args:
        session_id (str): _description_
        config_name (str): _description_
        rewrite (bool) : True if it needs to rewrite the existing json,
        multi_config (bool) : True if it has a multiple json configuration, then
        the configuration will be in the form of: [{}, {}, .. ]
    """
    session_root_dir = os.path.join(get_home_path(), CONFIG['REWARDS_PARENT_CONFIG_DIR'])

    configuration = [dict(kwargs)] if multi_config else dict(kwargs)
    save_path = os.path.join(session_root_dir, session_id, f"{config_name}.json")

    if rewrite and os.path.exists(save_path):
        existing_config = json.load(open(save_path, "r"))
        if multi_config:
            existing_config.append(configuration)
        else:
            existing_config[list(configuration.keys())[0]] = list(configuration.values())[0]
        configuration = existing_config

    with open(save_path, "w") as json_file:
        json.dump(configuration, json_file)


def add_dict_inside_session(
    session_id : str, configuration : dict, config_name : str, 
    session_root_folder : Optional[str] = None, rewrite : bool = False, multi_config : bool = False
):
    session_root_dir = os.path.join(
        get_home_path(), 
        CONFIG['REWARDS_PARENT_CONFIG_DIR']) if session_root_folder is None else os.path.join(
            session_root_folder, CONFIG['REWARDS_PARENT_CONFIG_DIR']
    )
    
    save_path = os.path.join(session_root_dir, session_id, f'{config_name}.json')
    
    if rewrite and os.path.exists(save_path):
        existing_config = json.load(open(save_path, "r"))
        if multi_config:
            existing_config.append(configuration)
        else:
            existing_config[list(configuration.keys())[0]] = list(configuration.values())[0]
        configuration = existing_config
    
    with open(save_path, "w") as json_file:
        json.dump(configuration, json_file)



def get_session_files(session_id: str, session_root_dir: Optional[str] = None) -> Dict[str, Any]:
    """Get all the files inside a session configuration folder.

    Args:
        session_id (str): The name of the session sub folder inside the root directory
        session_root_dir (Optional[str], optional): The directory where all the session config are been kept. Defaults to None.

    Returns:
        Dict[str, Any]: All the configurations for a particular session
    """
    session_root_dir = (
        os.path.join(get_home_path(), CONFIG['REWARDS_PARENT_CONFIG_DIR'])
        if session_root_dir is None
        else session_root_dir
    )

    session_dir = os.path.join(session_root_dir, session_id)
    all_configs = {}

    for json_file_name in os.listdir(session_dir):
        if json_file_name.endswith(".json") and not json_file_name.startswith("metrics"):
            json_dict = json.load(open(os.path.join(session_dir, json_file_name), "r"))
            all_configs[json_file_name.split(".json")[0]] = json_dict

    # TODO: Show the metrics and model configurations so that it can be displayed in fronend
    return all_configs


def get_all_sessions_info(session_root_dir : Optional[str] = None) -> Dict[str, Dict[str, Any]]:
    """
    Get all the sessions information present for a particular user. 
    
    Args:
        session_root_dir : Optional[str] -> The root directory where all the sessions are stored
    
    Returns:
        Dict[str, Dict[str, Any]]: All the configurations for all the sessions
    """
    all_session_infos = {}
    session_root_dir = (
        os.path.join(get_home_path(), CONFIG['REWARDS_PARENT_CONFIG_DIR'])
        if session_root_dir is None
        else session_root_dir
    )
    
    session_id_paths = [
        os.path.join(
            session_root_dir, session_id
        ) for session_id in os.listdir(session_root_dir)]
    
    for session_id, session_id_path in zip(os.listdir(session_root_dir), session_id_paths):
        all_session_infos[session_id] = get_session_files(session_id=session_id, session_root_dir=session_root_dir)
    
    return all_session_infos

def delete_session(session_id : str, session_root_dir : Optional[str] = None):
    """Deletes a particulart session 

    Args:
        session_id (str): The name of the session sub folder inside the root directory
        session_root_dir (Optional[str], optional): The directory where all the session config are been kept. Defaults to None.

    """
    session_root_dir = (
        os.path.join(get_home_path(), CONFIG['REWARDS_PARENT_CONFIG_DIR'])
        if session_root_dir is None
        else session_root_dir
    )
    session_dir = os.path.join(session_root_dir, session_id)
    try:
        shutil.rmtree(session_dir)
        return {
            'status' : 200, 
            'message': f'Session {session_dir} deleted succesfully'
        }
    except Exception as e:
        return {
            'status' : 500, 
            'message': f'Internal server error {e}'
        }


# TBD NOT DONE 

import matplotlib.pyplot as plt
import pygame
import time
import torch

plt.ion()

N_CARS = 1

plot_scores = []
plot_mean_scores = []
total_score = 0
record = 0

CARS = [Car()] * N_CARS
MODELS = [Linear_QNet([5, 128, 3])] * N_CARS
MODEL_PATHS = [r"model/model.pth"] * N_CARS

for i, model in enumerate(MODELS):
    model.load_state_dict(torch.load(MODEL_PATHS[i]))
    model.eval()

game = GameController(CARS)

while True:
    time.sleep(0.05)
    pygame.display.update()
    action = [[0]*5]*N_CARS

    res = [MODELS[i](torch.tensor(CARS[i].radars, dtype=torch.float)) for i in range(N_CARS)]

    for i, act in enumerate(action):
        print(torch.argmax(res[i]).item(), i)
        action[i][torch.argmax(res[i]).item()] = 1

    game.play_Step(action)
    if not game.alive:
        game.initialize()



import torch 
import random 
import numpy as np 
from pathlib import Path 
from collections import deque
from dataclasses import dataclass
from typing import Any, List, Optional, Union 


# TODO: random.randint -> random.uniform(0, 1) with a SEED 
# TODO: take the float value of random rather than converting it into int 
# TODO: Implement epsilon decay 

@dataclass
class AgentConf:
    MAX_MEMORY: int = 100000  
    BATCH_SIZE: int = 1000 
    PARENT_PATH: str = str(Path(__file__).parent.parent)
    DEVICE: str = "cpu" 
    TRACK: bool = False 

class Agent(AgentConf):
    def __init__(
        self,
        model: torch.nn.Module,
        checkpoint_folder_path: Optional[str] = None,
        model_name : Optional[str] = None, 
        lr: float = 0.01,
        epsilon: float = 0.25,
        gamma: float = 0.9,
    ) -> None:
        super(Agent, self).__init__()
        """The Agent class which acts as a RL agent similar like Open AI's gym agent

        Args:
            model (torch.nn.Module): _description_
            checkpoint_path (Optional[str], optional): The model checkpoint to load its weight. Defaults to False.
            lr (float, optional): _description_. Defaults to 0.01
            epsilon (float, optional): _description_. Defaults to 0.25.
            gamma (float, optional): _description_. Defaults to 0.9.
        """
        self.n_games = 0
        self.epsilon = epsilon * 100 if epsilon < 1 else epsilon
        self.lr = lr
        self.gamma = gamma

        self.memory = deque(maxlen=self.MAX_MEMORY)
        self.model = model
        
        # Lates changes loading the model directly if exists 
        self.model.load(checkpoint_folder_path, model_name, self.DEVICE)
        self.checkpoint_folder_path = checkpoint_folder_path
        self.model_name = model_name
    
    def get_state(self, game: Any) -> np.ndarray:
        """Returns the current state of the game.
        NOTE: Some Assumptions:
        - We assume that the game environment is made using pygame
        - We also assume that the agent inside the game uses `radars` that keeps track of its all position and other parameters.

        Args:
            game (rewards.env.car.CarGame): The current game environment of pygame.

        Returns:
            np.ndarray: An array of the state of the game. Here it is the agent's radars.
        """

        # TODO: Check the type of game
        state = game.radars
        return np.array(state, dtype=int)

    def remember(
        self,
        state: np.ndarray,
        action: Union[np.ndarray, List[int]],
        reward: Union[int, float],
        next_state: np.ndarray,
        done: bool,
    ) -> List[Union[float, int]]:
        """Remmembers the state of the game for the exploration phase of the agent

        Args:
            state (np.ndarray): The current state of the agent
            action (Union[np.ndarray, List[int]]): Action taken by the agent
            reward (Union[int, float]): Reward that the agent gets
            next_state (np.ndarray): The next state which the agent takes after taking the current action
            done (bool): Whether the game is finished or not.

        Returns:
            List[Union[float, int]]: The final move after exploration which is an action
        """
        self.memory.append((state, action, reward, next_state, done))
    
    def get_action(self, state, mode = "training"):
        """
        Two modes available:
        - training 
        - evaluation
        """
        final_move = [0, 0, 0]
        if mode == "training":
            if random.randint(0, 100) < self.epsilon:
                move = random.randint(0, 2)
                final_move[move] = 1
            else:          
                state0 = torch.tensor(state, dtype=torch.float)
                prediction = self.model(state0)
                move = torch.argmax(prediction).item()
                final_move[move] = 1
            return final_move
        
        else:
            state0 = torch.tensor(state, dtype=torch.float)
            prediction = self.model(state0)
            move = torch.argmax(prediction).item()
            final_move[move] = 1
            return final_move
        


import os
import glob
import datetime 
from typing import List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


DEFAULT_MODEL_PATH = './saved_models/'

class DeepNet(nn.Module):
    """
    Base Class for all the models that will be added from here and inherited from this class.
    """

    def __init__(self) -> None:
        super(DeepNet, self).__init__()

    def save(
        self, checkpoint_folder_path : Optional[str] = None, model_name : Optional[str] = None, device : Optional[str] = "cpu"
    ) -> None:
        """Loads the model in a robust file structure and setting 

        Args:
            checkpoint_folder_path (Optional[str], optional): The model checkpoint folder where all the checkpoints are been saved. Defaults to None.
            model_name (Optional[str], optional): The name of the model to save. Defaults to None.
        """
        folder_path = DEFAULT_MODEL_PATH if checkpoint_folder_path is None else checkpoint_folder_path
        _model_name = f'model_{datetime.datetime.now()}_.pth' if model_name is None else model_name
        
        if folder_path == DEFAULT_MODEL_PATH and not os.path.exists(DEFAULT_MODEL_PATH):
            os.mkdir(DEFAULT_MODEL_PATH)
            print("=> Creating DEFAULT MODEL FOLDER PATH to saave all the model checkpoints")
        
        torch.save(
            self.state_dict(), 
            os.path.join(folder_path, _model_name)
        )
        print(f"=> Latest model saved as {_model_name}")
    
    def _get_file_list_mod_by_date(self, search_dir : str, reversed : Optional[bool] = True) -> List[str]:
        """Lists the files based on the date modified 

        Args:
            search_dir (str): The folder to search and sort 
            reversed (Optional[bool], optional): Wants in descending order. Defaults to True.

        Returns:
            List[str]: The list of sorted files 
        """
        files = list(filter(os.path.isfile, glob.glob(search_dir + "*")))
        files.sort(key=lambda x: os.path.getmtime(x), reverse=reversed)
        return files 
    
    def load(self, checkpoint_folder_path : Optional[str] = None, model_name : Optional[str] = None, device : Optional[str] = "cpu") -> None:
        """Loads the model in a robust file structure and setting 

        Args:
            checkpoint_folder_path (Optional[str], optional): The model checkpoint folder where all the checkpoints are been saved. Defaults to None.
            model_name (Optional[str], optional): The name of the model to save. Defaults to None.
        """
        if checkpoint_folder_path and len(os.listdir(checkpoint_folder_path)) > 0:
            print(checkpoint_folder_path, self._get_file_list_mod_by_date(checkpoint_folder_path))

            _model_name = self._get_file_list_mod_by_date(checkpoint_folder_path)[0] if model_name is None else model_name 
            model_path = os.path.join(checkpoint_folder_path, _model_name)
            try:
                self.load_state_dict(
                    torch.load(
                        model_path,
                        map_location=device,
                    )
                )
                self.eval()
                print(f"=> Model loaded successfully from {model_path}")
                
            except Exception as e:
                # TODO: Write custom error message from exceptions and logging 
                print(f"=> Error occured {e}")
                
        # give the options to load from the default checkpoint folder path and files 
        # for this version the default checkpoint folder path will be ./saved_models and the latest model will be taken in consideration 
        
        elif os.path.exists(DEFAULT_MODEL_PATH):
            if len(os.listdir(DEFAULT_MODEL_PATH)) > 0 and model_name is None: 
                _default_model_path = self._get_file_list_mod_by_date(DEFAULT_MODEL_PATH)[0]
            
            else:
                _default_model_path = os.path.join(DEFAULT_MODEL_PATH, model_name)
            
            print(_default_model_path)
            try:
                self.load_state_dict(
                    torch.load(_default_model_path, map_location=device)
                )
                self.eval() 
                print(f"=> Found default location and loaded successfully from {_default_model_path}")
            except Exception as e:
                print(f"=> No model found  Exception occured at: {e}")
        else:
            print("=> No location or default location found")
        

class LinearQNet(DeepNet):
    def __init__(self, layers_conf: List[List[int]]):
        """Basic LinearQNet for the agent model.

        Args:
            layers_conf (List[List[int]]): The list of layers. Each element will be a list
            example: [[in_dim, out_dim], [in_dim, out_dim]]
        """
        super(LinearQNet, self).__init__()
        self.layers_conf = layers_conf
        self.num_layers = len(layers_conf)
        self.layers = nn.ModuleList()
        for i in range(self.num_layers):
            self.layers.append(nn.Linear(layers_conf[i][0], layers_conf[i][1]))

    def forward(self, x):
        for i in range(self.num_layers - 1):
            x = F.relu(self.layers[i](x))
        return self.layers[-1](x)



from rewards_experimental.config import CONFIG 
from rewards_experimental.agent import Agent 
from rewards_experimental.trainer import QTrainer 
from rewards_experimental.models import DeepNet, LinearQNet
from rewards_experimental.workflow import RewardsWorkflow

__version__ = "0.0.1"